{***************************************************}
{* Simulation routines for LogicGates9 application *}
{***************************************************}

unit common;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, Controls, StdCtrls, ExtCtrls, Graphics;

type
  // Switches
  TSwitch = record
    shSwitchH, shSwitchV: TShape;                                              // horizontal + vertical parts of switch shape
    imSwitch: TImage;                                                          // transparent image as switch click area
    iPosOffH, iPosOffV: Integer;                                               // off position (Top property) of horiz. + vert. switch shapes
    bStatus: Boolean;                                                          // switch on-off status
  end;
  // LEDs
  TLed = record
    shLed: TShape;                                                             // LED shape
    bStatus: Boolean;                                                          // LED on-off status
  end;
  // MOSFETs
  TMosfet = record
    laStatus: TLabel;                                                          // MOSFET status label
    cChannel: Char;                                                            // MOSFET channel (N or P)
    bStatus: Boolean;                                                          // MOSFET on-off status (= high-low output)
  end;

const
  // "Renaming" of True/False constants for better program readability
  SwitchOn = True; SwitchOff = False;
  High = True; Low = False;
  // UTF8 subscript codes
  SUB_1 = #$E2#$82#$81; SUB_2 = #$E2#$82#$82;
  SUB_3 = #$E2#$82#$83; SUB_4 = #$E2#$82#$84;

// Public procedures (these procedures contain all simulation code used in the other units)

procedure Init(var SwitchA, SwitchB: TSwitch; var LedA, LedB, LedOut: TLed; var MF1, MF2: TMosfet);
procedure Init2(var SwitchA, SwitchB: TSwitch; var LedA, LedB, LedOut: TLed; var MF1, MF2, MF3, MF4: TMosfet);
procedure GateNand1(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2: TMosfet; var LedOut: TLed);
procedure GateNor1(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2: TMosfet; var LedOut: TLed);
procedure GateNand2(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2, Mosfet3, Mosfet4: TMosfet; var LedOut: TLed);
procedure GateNor2(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2, Mosfet3, Mosfet4: TMosfet; var LedOut: TLed);
procedure SwitchPush(var Switch: TSwitch; var Led: TLed);
procedure SwitchOnOff(var Switch: TSwitch; Status: Boolean);
procedure LedOnOff(var Led: TLed; Status: Boolean);
procedure MosfetsSeries(Mosfet1Input, Mosfet2Input: Boolean; var Mosfet1, Mosfet2: TMosfet; out MosfetsOutput: Boolean);
procedure MosfetsParallel(Mosfet1Input, Mosfet2Input: Boolean; var Mosfet1, Mosfet2: TMosfet; out MosfetsOutput: Boolean);
procedure MosfetOnOff(var Mosfet: TMosfet; MosfetInput: Boolean);

implementation

{ Initialisation of NAND/NOR with resistor simulation}

procedure Init(var SwitchA, SwitchB: TSwitch; var LedA, LedB, LedOut: TLed; var MF1, MF2: TMosfet);

begin
  SwitchOnOff(SwitchA, SwitchOff); SwitchOnOff(SwitchB, SwitchOff);            // both switches off
  LedOnOff(LedA, Low); LedOnOff(LedB, Low);                                    // both input LEDs off
  LedOnOff(LedOut, High);                                                      // output LED on
  MosfetOnOff(MF1, Low); MosfetOnOff(MF2, Low);                                // both MOSFETs input = logic low
end;

{ Initialisation of CMOS NAND/NOR simulation}

procedure Init2(var SwitchA, SwitchB: TSwitch; var LedA, LedB, LedOut: TLed; var MF1, MF2, MF3, MF4: TMosfet);

begin
  Init(SwitchA, SwitchB, LedA, LedB, LedOut, MF1, MF2);                        // same initial values as for "with resistor" simulations
  MosfetOnOff(MF3, Low); MosfetOnOff(MF4, Low);                                // both "pull-up MOSFETs" input = logic low
end;

{ MOSFET and resistor NAND gate simulation }

procedure GateNand1(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2: TMosfet; var LedOut: TLed);

var
  Mosfet1Input, Mosfet2Input, MosfetsOutput, LedOutStatus: Boolean;

begin
  // Determine output of M1 and M2 in series circuit
  Mosfet1Input := SwitchA.bStatus; Mosfet2Input := SwitchB.bStatus;
  MosfetsSeries(Mosfet1Input, Mosfet2Input, Mosfet1, Mosfet2, MosfetsOutput);
  // Output LED is on (high) by default
  // If the output of M1 and M2 in series circuit is high, output LED will be out (low)
  LedOutStatus := High;
  if MosfetsOutput = High then
    LedOutStatus := Low;
  LedOnOff(LedOut, LedOutStatus);                                              // turn output LED on/off
end;

{ MOSFET and resistor NOR gate simulation }

procedure GateNor1(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2: TMosfet; var LedOut: TLed);

var
  Mosfet1Input, Mosfet2Input, MosfetsOutput, LedOutStatus: Boolean;

begin
  // Determine output of M1 and M2 in parallel circuit
  Mosfet1Input := SwitchA.bStatus; Mosfet2Input := SwitchB.bStatus;
  MosfetsParallel(Mosfet1Input, Mosfet2Input, Mosfet1, Mosfet2, MosfetsOutput);
  // Output LED is on (high) by default
  // If the output of M1 and M2 in paralell circuit is high, output LED will be out (low)
  LedOutStatus := High;
  if MosfetsOutput = High then
    LedOutStatus := Low;
  LedOnOff(LedOut, LedOutStatus);
end;

{ CMOS MOSFET NAND gate simulation }

procedure GateNand2(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2, Mosfet3, Mosfet4: TMosfet; var LedOut: TLed);

var
  Mosfet1Input, Mosfet2Input, Mosfet3Input, Mosfet4Input, Mosfets12Output, Mosfets34Output, LedOutStatus: Boolean;

begin
  // Determine output of M1 and M2 in series circuit
  Mosfet1Input := SwitchA.bStatus; Mosfet2Input := SwitchB.bStatus;
  MosfetsSeries(Mosfet1Input, Mosfet2Input, Mosfet1, Mosfet2, Mosfets12Output);
  // Determine output of M3 and M4 in parallel circuit
  Mosfet3Input := Mosfet1Input; Mosfet4Input := Mosfet2Input;
  MosfetsParallel(Mosfet3Input, Mosfet4Input, Mosfet3, Mosfet4, Mosfets34Output);
  // By default (A or B = low), output LED status depends on the output of M3 and M4 in parallel circuit (output LED = on)
  // If the output of M1 and M2 in series circuit is high, output LED will be out (low)
  LedOutStatus := Mosfets34Output;
  if Mosfets12Output = High then
    LedOutStatus := Low;
  LedOnOff(LedOut, LedOutStatus);
end;

{ CMOS MOSFET NOR gate simulation }

procedure GateNor2(SwitchA, SwitchB: TSwitch; var Mosfet1, Mosfet2, Mosfet3, Mosfet4: TMosfet; var LedOut: TLed);

var
  Mosfet1Input, Mosfet2Input, Mosfet3Input, Mosfet4Input, Mosfets12Output, Mosfets34Output, LedOutStatus: Boolean;

begin
  // Determine output of M1 and M2 in parallel circuit
  Mosfet1Input := SwitchA.bStatus; Mosfet2Input := SwitchB.bStatus;
  MosfetsParallel(Mosfet1Input, Mosfet2Input, Mosfet1, Mosfet2, Mosfets12Output);
  // Determine output of M3 and M4 in series circuit
  Mosfet3Input := Mosfet1Input; Mosfet4Input := Mosfet2Input;
  MosfetsSeries(Mosfet3Input, Mosfet4Input, Mosfet3, Mosfet4, Mosfets34Output);
  // By default (A and B = low), output LED status depends on the output of M3 and M4 in series circuit (output LED = on)
  // If the output of M1 and M2 in parallel circuit is high, output LED will be out (low)
  LedOutStatus := Mosfets34Output;
  if Mosfets12Output = High then
    LedOutStatus := Low;
  LedOnOff(LedOut, LedOutStatus);
end;

{ Switch A resp. B pushed: Adapt switch status and position and corr. LED status and display}

procedure SwitchPush(var Switch: TSwitch; var Led: TLed);

// This procedure concerns only the switch (and its LED), not the gate simulation

var
  Status: Boolean;

begin
  Status := not Switch.bStatus;                                                // change switch on-off status
  SwitchOnOff(Switch, Status);                                                 // adapt switch status and position
  LedOnOff(Led, Status);                                                       // adapt corr. LED status and display
end;

{ Set switch status and position }

procedure SwitchOnOff(var Switch: TSwitch; Status: Boolean);

begin
  Switch.bStatus := Status;
  if Switch.bStatus = SwitchOff then begin
    // Switch is in "off" position: top of shapes is default position
    Switch.shSwitchH.Top := Switch.iPosOffH;
    Switch.shSwitchV.Top := Switch.iPosOffV;
    Switch.imSwitch.Top := Switch.shSwitchV.Top - 2;
  end
  else begin
    // Switch is in "on" position: top of shapes 10 pixels down relative to default position
    Switch.shSwitchH.Top := Switch.iPosOffH + 10;
    Switch.shSwitchV.Top := Switch.iPosOffV + 10;
    Switch.imSwitch.Top := Switch.shSwitchV.Top - 2;
  end;
end;

{ Set LED status and display }

procedure LedOnOff(var Led: TLed; Status: Boolean);

begin
  Led.bStatus := Status;
  if Led.bStatus = High then begin
    // LED is on: solid color display
    Led.shLed.Brush.Style := bsSolid;
  end
  else begin
    // LED is off: pattern color display
    Led.shLed.Brush.Style := bsDiagCross;
  end;
end;

{ MOSFETs in series circuit }

procedure MosfetsSeries(Mosfet1Input, Mosfet2Input: Boolean; var Mosfet1, Mosfet2: TMosfet; out MosfetsOutput: Boolean);

begin
  // Determine first MOSFET's output
  MosfetOnOff(Mosfet1, Mosfet1Input);
  // Determine second MOSFET's output
  MosfetOnOff(Mosfet2, Mosfet2Input);
  // Series circuit output behaves like a logic AND
  MosfetsOutput := Mosfet1.bStatus and Mosfet2.bStatus;
end;

{ MOSFETs in series circuit }

procedure MosfetsParallel(Mosfet1Input, Mosfet2Input: Boolean; var Mosfet1, Mosfet2: TMosfet; out MosfetsOutput: Boolean);

begin
  // Determine first MOSFET's output
  MosfetOnOff(Mosfet1, Mosfet1Input);
  // Determine second MOSFET's output
  MosfetOnOff(Mosfet2, Mosfet2Input);
  // Parallel circuit output behaves like a logic OR
  MosfetsOutput := Mosfet1.bStatus or Mosfet2.bStatus;
end;

{ Determine MOSFET output and adapt on-off status display }

procedure MosfetOnOff(var Mosfet: TMosfet; MosfetInput: Boolean);

begin
  // MOSFET logic output depends on channel type
  if Mosfet.cChannel = 'N' then
    Mosfet.bStatus := MosfetInput
  else
    Mosfet.bStatus := not MosfetInput;
  // Adapt MOSFET on-off status display
  if Mosfet.bStatus = High then begin
    // MOSFET is on (output = high)
    Mosfet.laStatus.Caption := 'on';
    Mosfet.laStatus.Font.Color := clLime;
  end
  else begin
    // MOSFET is off (output = low)
    Mosfet.laStatus.Caption := 'off';
    Mosfet.laStatus.Font.Color := clRed;
  end;
end;

end.

